/*
 * Calibrate Pololu MinIMU-9 measurements.
 *
 * Copyright (C) 2013 Paul Boddie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include "imu.h"
#include "shutdown.h"
#include "geo.h"

vectorf vmin = {{0, 0, 0}},
        vmax = {{0, 0, 0}};

void calibrate_shutdown(int signum)
{
    printf("\n%-6.1f %-6.1f %-6.1f %-6.1f %-6.1f %-6.1f\n",
        vmin.x, vmin.y, vmin.z, vmax.x, vmax.y, vmax.z);
    init_shutdown(signum);
}

/* Main program. */

int main(int argc, char *argv[])
{
    uint8_t result[6];
    vectorf value,
            valueB[IMU_ACCEL_BUFFER_SIZE];
    int argno = 1, vindex = 0;
    bool gyroscope = false, accelerometer = false, magnetometer = false,
         averaging = false, normalised = false;

    if ((argc > argno) && (strcmp(argv[argno], "-g") == 0))
    {
        argno++;
        gyroscope = true;
    }
    else if ((argc > argno) && (strcmp(argv[argno], "-a") == 0))
    {
        argno++;
        accelerometer = true;
    }
    else if ((argc > argno) && (strcmp(argv[argno], "-m") == 0))
    {
        argno++;
        magnetometer = true;
    }
    else
    {
        printf("Need -g, -a or -m to select gyroscope, accelerometer or magnetometer respectively.\n");
        exit(1);
    }

    if ((argc > argno) && (strcmp(argv[argno], "-v") == 0))
    {
        argno++;
        averaging = true;
    }

    if ((argc > argno) && (strcmp(argv[argno], "-n") == 0))
    {
        argno++;
        normalised = true;
    }

    signal(SIGINT, calibrate_shutdown);

    /* Access the 8:10 port. */

    if (ubb_open(0) < 0) {
        perror("ubb_open");
        return 1;
    }

    ubb_power(1);
    printf("Power on.\n");

    /* Bring the IMU up. */

    imu_init();

    if (gyroscope)
    {
        imu_sendone(IMU_GYRO_ADDRESS, IMU_GYRO_CTRL_REG1, IMU_GYRO_CTRL_REG1_ALL);
        imu_sendone(IMU_GYRO_ADDRESS, IMU_GYRO_CTRL_REG5, 0);
        imu_sendone(IMU_GYRO_ADDRESS, IMU_GYRO_CTRL_REG4, IMU_GYRO_CTRL_REG4_BDU | IMU_GYRO_DPS_SCALE);
    }
    else if (accelerometer)
    {
        imu_sendone(IMU_ACCEL_ADDRESS, IMU_ACCEL_CTRL_REG1_A, IMU_ACCEL_CTRL_REG1_ALL | IMU_ACCEL_CTRL_REG1_50HZ);

        if ((argc > argno) && (strcmp(argv[argno], "-f") == 0))
        {
            argno++;
            imu_sendone(IMU_ACCEL_ADDRESS, IMU_ACCEL_CTRL_REG2_A, IMU_ACCEL_CTRL_REG2_FDS | IMU_ACCEL_FILTER_FREQ);
            imu_recv(IMU_ACCEL_ADDRESS, IMU_ACCEL_HP_FILTER_RESET_A, result, 1);
        }
        else
            imu_sendone(IMU_ACCEL_ADDRESS, IMU_ACCEL_CTRL_REG2_A, 0);

        imu_sendone(IMU_ACCEL_ADDRESS, IMU_ACCEL_CTRL_REG4_A, IMU_ACCEL_CTRL_REG4_BDU | IMU_ACCEL_SCALE);
    }
    else if (magnetometer)
    {
        imu_sendone(IMU_MAGNET_ADDRESS, IMU_MAGNET_CRA_REG_M, IMU_MAGNET_CRA_REG_30HZ);
        imu_sendone(IMU_MAGNET_ADDRESS, IMU_MAGNET_CRB_REG_M, IMU_MAGNET_CRB_REG_4_0G);
        imu_sendone(IMU_MAGNET_ADDRESS, IMU_MAGNET_MR_REG_M, IMU_MAGNET_MR_REG_CONT);
    }

    if (imu_recv(IMU_GYRO_ADDRESS, IMU_GYRO_WHO_AM_I, result, 1))
        printf("Who am I? %x\n", result[0]);

    if (imu_recv(IMU_MAGNET_ADDRESS, IMU_MAGNET_WHO_AM_I_M, result, 1))
        printf("Who am I? %x\n", result[0]);

    if (imu_recv(IMU_MAGNET_ADDRESS, IMU_MAGNET_IRA_REG_M, result, 1))
        printf("Identification A? %x\n", result[0]);

    if (imu_recv(IMU_MAGNET_ADDRESS, IMU_MAGNET_IRB_REG_M, result, 1))
        printf("Identification B? %x\n", result[0]);

    if (imu_recv(IMU_MAGNET_ADDRESS, IMU_MAGNET_IRC_REG_M, result, 1))
        printf("Identification C? %x\n", result[0]);

    memset(valueB, 0, sizeof(valueB));

    while (1)
    {
        if (accelerometer)
            imu_read_vector(IMU_ACCEL_ADDRESS, IMU_ACCEL_OUT_X_L_A | IMU_ACCEL_READ_MANY, &value, convert12);
        else if (gyroscope)
            imu_read_vector(IMU_GYRO_ADDRESS, IMU_GYRO_OUT_X_L | IMU_GYRO_READ_MANY, &value, convert);
        else if (magnetometer)
            imu_read_vector_xzy(IMU_MAGNET_ADDRESS, IMU_MAGNET_OUT_X_H_M, &value, convertBE12L);

        if (averaging)
        {
            queue(valueB, &vindex, IMU_ACCEL_BUFFER_SIZE, &value);
            filter(valueB, vindex, IMU_ACCEL_BUFFER_SIZE, &value);
        }

        vmin.x = min(value.x, vmin.x);
        vmin.y = min(value.y, vmin.y);
        vmin.z = min(value.z, vmin.z);
        vmax.x = max(value.x, vmax.x);
        vmax.y = max(value.y, vmax.y);
        vmax.z = max(value.z, vmax.z);

        printf("V: % 6.1f, % 6.1f, % 6.1f  "
               "-: % 6.1f, % 6.1f, % 6.1f  "
               "+: % 6.1f, % 6.1f, % 6.1f\n",
               normalised ? scale(value.x, vmin.x, (vmin.x + vmax.x) / 2, vmax.x) : value.x,
               normalised ? scale(value.y, vmin.y, (vmin.y + vmax.y) / 2, vmax.y) : value.y,
               normalised ? scale(value.z, vmin.z, (vmin.z + vmax.z) / 2, vmax.z) : value.z,
               vmin.x, vmin.y, vmin.z,
               vmax.x, vmax.y, vmax.z);

        usleep(magnetometer ? IMU_MAGNET_UPDATE_PERIOD : IMU_UPDATE_PERIOD);
    }

    /* This should be unreachable. */

    ubb_close(0);
    return 0;
}
