/*
 * Geometric operations.
 *
 * Copyright (C) 2013 Paul Boddie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include "geo.h"

/**
 * Return a value within the given magnitude.
 */
double within(double value, double mag)
{
    double offset = fmod(value, mag);
    int num = (int) fdiv(value, mag);

    if (num % 2 == 0)
        return offset;
    else if (offset > 0)
        return offset - mag;
    else
        return offset + mag;
}

/**
 * Return whether the given vector is null.
 */
bool vectorf_null(const vectorf *in)
{
    return ((in->x == 0) && (in->y == 0) && (in->z == 0));
}

/**
 * Reset the given vector to have zero magnitude.
 */
void vectorf_reset(vectorf *in)
{
    in->x = 0; in->y = 0; in->z = 0;
}

/**
 * Return the cross product of the two input vectors in the output vector
 * provided.
 */
void vectorf_cross(const vectorf *in1, const vectorf *in2, vectorf *out)
{
    out->x = in1->y * in2->z - in1->z * in2->y;
    out->y = in1->z * in2->x - in1->x * in2->z;
    out->z = in1->x * in2->y - in1->y * in2->x;
}

/**
 * Return the dot product of the two input vectors.
 */
double vectorf_dot(const vectorf *in1, const vectorf *in2)
{
    return in1->x * in2->x + in1->y * in2->y + in1->z * in2->z;
}

/**
 * Return the magnitude of the given vector.
 */
double vectorf_mag(const vectorf *in)
{
    return sqrt(pow(in->x, 2) + pow(in->y, 2) + pow(in->z, 2));
}

/**
 * Normalise the given vector.
 */
void vectorf_normalise(vectorf *in, vectorf *out)
{
    double mag = vectorf_mag(in);

    out->x = in->x / mag;
    out->y = in->y / mag;
    out->z = in->z / mag;
}

/**
 * Rotate the plane defined by the given input vectors by the given angle,
 * in radians, modifying the input vectors to define the resulting plane.
 */
void plane_rotate(vectorf *in1, vectorf *in2, double angle)
{
    double c = cos(angle), s = sin(angle);
    vectorf out1, out2;

    /* Rotation around the plane normal with contributions from both plane
       vectors. */

    out1.x = c * in1->x + s * in2->x;
    out1.y = c * in1->y + s * in2->y;
    out1.z = c * in1->z + s * in2->z;

    /* The second plane vector is 90 degrees (pi/2) ahead of the first:
       cos(angle + pi/2) == -sin(angle)
       sin(angle + pi/2) == cos(angle) */

    out2.x = c * in2->x - s * in1->x;
    out2.y = c * in2->y - s * in1->y;
    out2.z = c * in2->z - s * in1->z;

    *in1 = out1;
    *in2 = out2;
}

/**
 * Populate the output vector with the negated input vector.
 */
void vectorf_negate(vectorf *in, vectorf *out)
{
    out->x = -in->x;
    out->y = -in->y;
    out->z = -in->z;
}

/**
 * Return the direction of the given vector in radians.
 */
double vectorf_direction(vectorf *in)
{
    return atan2(in->x, in->z);
}

/**
 * Return the elevation of the given unit vector in radians.
 */
double vectorf_elevation(vectorf *in)
{
    return asin(in->y);
}

/**
 * Obtain the direction vector for the given direction and elevation.
 */
void vectorf_polar(double direction, double elevation, vectorf *out)
{
    out->x = sin(direction) * cos(elevation);
    out->y = sin(elevation);
    out->z = cos(direction) * cos(elevation);
}

/**
 * Convert the input vector expressed in terms of the given axis vectors into
 * the global coordinate space.
 */
void vectorf_convert(const vectorf *in, const vectorf *devicex, const vectorf *devicey, const vectorf *devicez, vectorf *out)
{
    out->x = in->x * devicex->x + in->y * devicey->x + in->z * devicez->x;
    out->y = in->x * devicex->y + in->y * devicey->y + in->z * devicez->y;
    out->z = in->x * devicex->z + in->y * devicey->z + in->z * devicez->z;
}

/**
 * Convert the input vector expressed in terms of the global coordinate space
 * into a vector expressed in terms of the given axis vectors.
 */
void vectorf_convert_into(const vectorf *in, const vectorf *x, const vectorf *y, const vectorf *z, vectorf *out)
{
    out->x = vectorf_dot(in, x);
    out->y = vectorf_dot(in, y);
    out->z = vectorf_dot(in, z);
}

/**
 * Return the tilt/roll of the given vector in radians projected into the given
 * plane.
 */
double vectorf_tilt_in_plane(const vectorf *in, const vectorf *x, const vectorf *y)
{
    vectorf v, z, y2;

    vectorf_cross(x, y, &z);
    vectorf_normalise(&z, &z);
    vectorf_cross(&z, x, &y2);
    vectorf_convert_into(in, x, &y2, &z, &v);
    return atan2(v.y, v.x);
}

/**
 * Return the tilt/roll of the given vector in radians projected into the given
 * space.
 */
double vectorf_tilt_in_plane_with_axis(const vectorf *in, const vectorf *x, const vectorf *y, const vectorf *z)
{
    vectorf v, z2, y2;

    vectorf_cross(x, y, &z2);
    if (vectorf_dot(z, &z2) < 0)
        vectorf_negate(&z2, &z2);
    vectorf_cross(&z2, x, &y2);
    vectorf_convert_into(in, x, &y2, &z2, &v);
    return atan2(v.y, v.x);
}

/**
 * Rotate a vector around the given z axis by the given angle in radians.
 */
void vectorf_rotate_in_space(const vectorf *in, const vectorf *x, const vectorf *y, const vectorf *z, double angle, vectorf *out)
{
    vectorf v;
    double tilt, c, s, vz, mag;

    /* Project the vector into the xy plane. */

    vectorf_convert_into(in, x, y, z, &v);

    /* Remember the z value before obtaining a two dimensional vector. */

    vz = v.z;
    v.z = 0;
    mag = vectorf_mag(&v);

    /* Get the current angle. */

    tilt = atan2(v.y, v.x);

    /* Calculate the rotated vector. */

    c = cos(tilt + angle);
    s = sin(tilt + angle);

    out->x = mag * (c * x->x + s * y->x) + vz * z->x;
    out->y = mag * (c * x->y + s * y->y) + vz * z->y;
    out->z = mag * (c * x->z + s * y->z) + vz * z->z;
}
