/*
 * Ben NanoNote graphical user interface utilities.
 *
 * Copyright (C) 2013 Paul Boddie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <stdio.h>
#include "gui.h"

SDL_Surface *screen;
gui_printer printer = {0, 0, 255, 255, 255, 255};

void gui_init()
{
    if (SDL_Init(SDL_INIT_VIDEO) < 0)
    {
        fprintf(stderr, "SDL error: %s\n", SDL_GetError());
        exit(1);
    }
}

void gui_display_init()
{
    screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, 0, 0);
    SDL_ShowCursor(SDL_DISABLE);
}

void gui_shutdown(int signum)
{
    SDL_Quit();
    text_shutdown(signum);
}

void gui_shutdown_threaded(int signum)
{
    SDL_Quit();
    text_shutdown_threaded(signum);
}

void gui_quit()
{
    SDL_Quit();
    text_quit();
}

void gui_clear()
{
    gui_fill(0, 0, 0);
    printer.x = 0;
    printer.y = 0;
}

void gui_fill(uint8_t r, uint8_t g, uint8_t b)
{
    SDL_FillRect(screen, NULL, SDL_MapRGB(screen->format, r, g, b));
}

void gui_next_row(gui_printer *printer, uint16_t rows)
{
    printer->x = 0;
    printer->y = (printer->y + rows * SCREEN_ROW_HEIGHT) % SCREEN_HEIGHT;
}

void gui_next_column(gui_printer *printer, uint16_t columns)
{
    printer->x += columns * SCREEN_COLUMN_WIDTH;
    if (printer->x >= SCREEN_WIDTH)
    {
        printer->x = 0;
        printer->y = (printer->y + SCREEN_ROW_HEIGHT) % SCREEN_HEIGHT;
    }
}

int gui_printf(const char *format, ...)
{
    va_list args;
    static char buffer[GUI_BUFSIZE];
    char *this = buffer, *next;
    int printed = 0;

    va_start(args, format);
    vsnprintf(buffer, GUI_BUFSIZE, format, args);
    va_end(args);

    do
    {
        next = strchr(this, (int) '\n');
        if (next != NULL)
            *next = '\0';
        stringRGBA(screen, printer.x, printer.y, this, printer.r, printer.g, printer.b, printer.a);
        if (next != NULL)
            gui_next_row(&printer, 1);
        else
            gui_next_column(&printer, strlen(this));
        printed += strlen(this);
        this = next;
    }
    while (next != NULL);

    return printed;
}

/**
 * Generate sky points.
 */
void gui_sky(vectorf *viewx, vectorf *viewy, vectorf *viewz)
{
    double direction, elevation;
    int d, e, eclosest, estart, efinish, dcount;
    int16_t lastminx, lastminy, thisminx, thisminy,
            lastmaxx, lastmaxy, thismaxx, thismaxy;
    uint32_t colour;
    vectorf point;
    char buffer[5];

    direction = vectorf_direction(viewz);
    elevation = vectorf_elevation(viewz);

    d = closest(raddeg(direction), SKY_GRID_STEP);
    eclosest = closest(raddeg(elevation), SKY_GRID_STEP);

    /* Plot points between elevations -90 and 90 degrees. */

    estart = max(-90, eclosest - (SKY_GRID_STEP * 3));
    efinish = min(90, eclosest + (SKY_GRID_STEP * 3));

    for (e = estart; e <= efinish; e += SKY_GRID_STEP)
    {
        /* Start from the most central point and work outwards until
           the points are no longer visible. */

        for (dcount = 0; dcount <= 180; dcount += SKY_GRID_STEP)
        {
            /* Check point visibility. */

            vectorf_polar(degrad(d - dcount), degrad(e), &point);
            if (vectorf_dot(viewz, &point) <= 0)
                break;

            /* Plot points at the given extremes. */

            colour = e < 0 ? SKY_LOWER_COLOUR : SKY_UPPER_COLOUR;

            /* Current minimum extent of the direction. */

            thisminx = vectorf_dot(viewx, &point) * PROJECTION_FACTOR + SCREEN_WIDTH / 2;
            thisminy = vectorf_dot(viewy, &point) * -PROJECTION_FACTOR + SCREEN_HEIGHT / 2;

            /* Draw from the last point at this elevation. */

            if (dcount != 0)
            {
                lineColor(screen,
                    lastminx, lastminy,
                    thisminx, thisminy,
                    colour
                    );
            }

            /* Draw to the next elevation. */

            gui_sky_vertical(viewx, viewy, viewz, d - dcount, e, thisminx, thisminy, colour);

            /* Current maximum extent of the direction. */

            vectorf_polar(degrad(d + dcount), degrad(e), &point);
            thismaxx = vectorf_dot(viewx, &point) * PROJECTION_FACTOR + SCREEN_WIDTH / 2;
            thismaxy = vectorf_dot(viewy, &point) * -PROJECTION_FACTOR + SCREEN_HEIGHT / 2;

            /* Draw from the last point at this elevation. */

            if (dcount != 0)
            {
                lineColor(screen,
                    lastmaxx, lastmaxy,
                    thismaxx, thismaxy,
                    colour
                    );
            }

            /* Draw to the next elevation. */

            gui_sky_vertical(viewx, viewy, viewz, d + dcount, e, thismaxx, thismaxy, colour);

            /* Stop after one point at -90 or 90 degree elevations. */

            if (abs(e) == 90)
            {
                pixelColor(screen,
                    thisminx, thisminy,
                    colour
                    );
                break;
            }

            /* Write labels for the most central set of points. */

            if (e == eclosest)
            {
                sprintf(buffer, "%d", d - dcount);
                stringRGBA(screen, thisminx, thisminy, buffer, 255, 255, 255, 255);
                sprintf(buffer, "%d", d + dcount);
                stringRGBA(screen, thismaxx, thismaxy, buffer, 255, 255, 255, 255);
            }

            lastminx = thisminx; lastminy = thisminy;
            lastmaxx = thismaxx; lastmaxy = thismaxy;
        }
    }
}

void gui_sky_vertical(vectorf *viewx, vectorf *viewy, vectorf *viewz, int direction, int elevation, int16_t x, int16_t y, uint32_t colour)
{
    vectorf point;

    if ((elevation >= -(90 - SKY_GRID_STEP)) || (elevation <= (90 - SKY_GRID_STEP)))
    {
        vectorf_polar(degrad(direction), degrad(elevation + SKY_GRID_STEP), &point);
        lineColor(screen,
            x, y,
            vectorf_dot(viewx, &point) * PROJECTION_FACTOR + SCREEN_WIDTH / 2,
            vectorf_dot(viewy, &point) * -PROJECTION_FACTOR + SCREEN_HEIGHT / 2,
            colour
            );

        if (elevation == -(90 - SKY_GRID_STEP))
        {
            vectorf_polar(degrad(direction), degrad(elevation - SKY_GRID_STEP), &point);
            lineColor(screen,
                x, y,
                vectorf_dot(viewx, &point) * PROJECTION_FACTOR + SCREEN_WIDTH / 2,
                vectorf_dot(viewy, &point) * -PROJECTION_FACTOR + SCREEN_HEIGHT / 2,
                colour
                );
        }
    }
}

/**
 * Generate motion vector.
 */
void gui_motion(vectorf *viewx, vectorf *viewy, vectorf *viewz, vectorf *accelerationD)
{
    vectorf point;
    uint32_t colour;

    if (vectorf_dot(viewz, accelerationD) <= 0)
    {
        colour = MOTION_REVERSE_COLOUR;
        vectorf_negate(accelerationD, &point);
    }
    else
    {
        colour = MOTION_FORWARD_COLOUR;
        point = *accelerationD;
    }

    lineColor(screen,
        SCREEN_WIDTH / 2,
        SCREEN_HEIGHT / 2,
        vectorf_dot(viewx, &point) * PROJECTION_FACTOR + SCREEN_WIDTH / 2,
        vectorf_dot(viewy, &point) * -PROJECTION_FACTOR + SCREEN_HEIGHT / 2,
        colour
        );
}

/**
 * Show the given point in space relative to the origin.
 */
void gui_point(vectorf *viewx, vectorf *viewy, vectorf *viewz, vectorf *point, uint8_t r, uint8_t g, uint8_t b, uint8_t a)
{
    int16_t x, y, radius;
    double z;

    if (fabs(vectorf_dot(viewz, point)) < 1)
        return;

    z = vectorf_dot(viewz, point);
    x = vectorf_dot(viewx, point) * PROJECTION_FACTOR / z + SCREEN_WIDTH / 2;
    y = vectorf_dot(viewy, point) * -PROJECTION_FACTOR / z + SCREEN_HEIGHT / 2;
    radius = abs(PROJECTION_FACTOR / z);

    if (vectorf_dot(viewz, point) < 0)
        circleRGBA(screen,
            x, y, radius,
            r, g, b, a
            );
    else
        filledCircleRGBA(screen,
            x, y, radius,
            r, g, b, a
            );
}

/**
 * Generate a view from the given point in space of the origin.
 */
void gui_origin(vectorf *viewx, vectorf *viewy, vectorf *viewz, vectorf *point)
{
    vectorf origin;

    vectorf_negate(point, &origin);
    gui_point(viewx, viewy, viewz, &origin, 255, 255, 255, 127);
}

/**
 * Draw a simple bar representation of the given vector.
 */
void gui_bar(vectorf *value)
{
    lineRGBA(screen,
        SCREEN_WIDTH / 2,
        SCREEN_HEIGHT / 2,
        SCREEN_WIDTH / 2 + value->x * SCREEN_WIDTH / 2,
        SCREEN_HEIGHT / 2 - value->y * SCREEN_HEIGHT / 2,
        255, 255, 255, 255);

    circleRGBA(screen,
        SCREEN_WIDTH / 2 + value->x * SCREEN_WIDTH / 2,
        SCREEN_HEIGHT / 2 - value->y * SCREEN_HEIGHT / 2,
        fabs(value->z) * SCREEN_WIDTH / 2,
        255, 255, 255, 255);
}

void gui_plot(uint16_t x, uint16_t y, uint8_t r, uint8_t g, uint8_t b, uint8_t a)
{
    pixelRGBA(screen, x, y, r, g, b, a);
}

void gui_flush()
{
    SDL_Flip(screen);
}

imu_ui_op gui_handle_events()
{
    SDL_Event event;

    while (SDL_PollEvent(&event))
    {
        switch (event.type)
        {
            case SDL_KEYDOWN:
            switch (event.key.keysym.sym)
            {
                case SDLK_ESCAPE:
                case SDLK_q:
                return IMU_UI_OP_QUIT;

                case SDLK_c:
                return IMU_UI_OP_CALIBRATE;

                case SDLK_r:
                return IMU_UI_OP_RESET;

                case SDLK_p:
                return IMU_UI_OP_PAUSE;

                default:
                break;
            }
            break;

            case SDL_QUIT:
            return IMU_UI_OP_QUIT;

            default:
            break;
        }
    }

    return IMU_UI_OP_NULL;
}
