/*
 * Ben NanoNote I2C communication.
 *
 * Copyright (C) 2013 Paul Boddie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <time.h>
#include "i2c.h"

extern uint32_t I2C_SCL, I2C_SDA;

void wait()
{
    /* NOTE: Would want 600ns or 1300ns, but nanosleep appears too slow.
    struct timespec ts = {0, 600};
    nanosleep(&ts, NULL); */

    uint8_t i;
    for (i = 0; i < 200; i++);
}

/**
 * Declare the pins for initial I2C communications.
 */
void i2c_init()
{
    OUT(I2C_SCL);
    OUT(I2C_SDA);
}

/**
 * Initiate an I2C transaction.
 */
void i2c_start()
{
    OUT(I2C_SDA);
    CLR(I2C_SCL);
    wait();
    SET(I2C_SDA);
    SET(I2C_SCL);
    wait();
    CLR(I2C_SDA);
}

/**
 * Terminate an I2C transaction.
 */
void i2c_stop()
{
    OUT(I2C_SDA);
    CLR(I2C_SCL);
    wait();
    CLR(I2C_SDA);
    SET(I2C_SCL);
    wait();
    SET(I2C_SDA);
}

/**
 * Send an I2C acknowledgement to a transmitting device.
 */
void i2c_ack(bool ack)
{
    OUT(I2C_SDA);

    if (ack)
        CLR(I2C_SDA);
    else
        SET(I2C_SDA);

    SET(I2C_SCL);
    wait();
    CLR(I2C_SCL);

    IN(I2C_SDA);
}

/**
 * Receive a single byte from an I2C device as part of a transaction.
 */
uint8_t i2c_recv()
{
    uint8_t mask, result = 0;

    IN(I2C_SDA);
    CLR(I2C_SCL);

    for (mask = 0x80; mask; mask >>= 1)
    {
        SET(I2C_SCL);
        wait();

        if (PIN(I2C_SDA))
            result |= mask;

        CLR(I2C_SCL);
        wait();
    }

    return result;
}

/**
 * Receive into a buffer a transmission of the given length in bytes.
 */
void i2c_recvmany(uint8_t *data, uint8_t len)
{
    uint8_t *end = data + len;

    for (; data != end; data++, len--)
    {
        *data = i2c_recv();
        i2c_ack(len > 1);
    }
}

void i2c_wait()
{
    IN(I2C_SCL);
    while (!PIN(I2C_SCL));
    OUT(I2C_SCL);
}

/**
 * Send a single byte of data to an I2C device as part of a transaction,
 * returning whether the transmission succeeded.
 */
bool i2c_send(uint8_t data)
{
    uint8_t mask;
    bool status;

    OUT(I2C_SDA);
    CLR(I2C_SCL);

    for (mask = 0x80; mask; mask >>= 1)
    {
        wait();

        if (data & mask)
            SET(I2C_SDA);
        else
            CLR(I2C_SDA);

        SET(I2C_SCL);
        wait();
        CLR(I2C_SCL);
    }

    /* Wait for acknowledgement, failing if none is given. */

    IN(I2C_SDA);
    SET(I2C_SCL);
    wait();

    status = PIN(I2C_SDA);
    CLR(I2C_SCL);
    return !status;
}

/**
 * Send from the buffer provided a transmission with the given length to an I2C
 * device.
 */
bool i2c_sendmany(uint8_t *data, uint8_t len)
{
    uint8_t *end = data + len;

    for (; data != end; data++)
    {
        if (!i2c_send(*data))
            return false;

        /* NOTE: Should test for the slave holding the clock signal low. */
    }

    return true;
}
