/*
 * Ben NanoNote communication with the Pololu MinIMU-9 with the L3G4200D 3-axis
 * gyroscope and LSM303DLM accelerometer/magnetometer.
 *
 * http://www.pololu.com/catalog/product/1265
 *
 * Copyright (C) 2013 Paul Boddie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <string.h>
#include <unistd.h>
#include "imu.h"

const uint32_t I2C_SCL = IMU_SCL, I2C_SDA = IMU_SDA;

/**
 * Receive data from the indicated device, using the given device register, and
 * storing the result in the buffer provided, expecting a transmission of the
 * specified length in bytes. Return whether the transaction was successful.
 */
bool imu_recv(uint8_t device, uint8_t reg, uint8_t *result, uint8_t len)
{
    uint8_t data[] = {device | I2C_WRITE, reg};

    memset(result, 0, len);

    /* Select the register. */

    i2c_start();
    if (!i2c_sendmany(data, 2))
    {
        i2c_stop();
        return false;
    }

    /* Continue with a read from the device. */

    i2c_repeated_start();
    if (!i2c_send(device | I2C_READ))
    {
        i2c_stop();
        return false;
    }
    i2c_recvmany(result, len);

    /* Stop the communication. */

    i2c_stop();
    return true;
}

/**
 * Send a single byte of data to the indicated device, using the given device
 * register. Return whether the transaction was successful.
 */
bool imu_sendone(uint8_t device, uint8_t reg, uint8_t value)
{
    uint8_t data[] = {device | I2C_WRITE, reg, value};

    /* Select the register. */

    i2c_start();
    if (!i2c_sendmany(data, 3))
    {
        i2c_stop();
        return false;
    }

    /* Stop the communication. */

    i2c_stop();
    return true;
}

/**
 * Convert the first two bytes of the given raw data to a signed integer.
 */
int16_t convert(uint8_t raw[])
{
    uint16_t raw16 = raw[0] | raw[1] << 8;
    if (raw16 & 0x8000)
        return -(raw16 ^ 0xffff) - 1;
    else
        return raw16;
}

/**
 * Convert the first two bytes of the given raw data to a signed integer,
 * discarding the least significant four bits of the raw data.
 */
int16_t convert12(uint8_t raw[])
{
    uint16_t raw16 = ((raw[0] >> 4) & 0x0f) | (raw[1] << 4);
    if (raw16 & 0x0800)
        return -(raw16 ^ 0x0fff) - 1;
    else
        return raw16;
}

/**
 * Convert the first two bytes of the given raw data to a signed integer,
 * discarding the most significant four bits of the raw data.
 */
int16_t convertBE12L(uint8_t raw[])
{
    uint16_t raw16 = ((raw[0] & 0x0f) << 8) | raw[1];
    if (raw16 & 0x0800)
        return -(raw16 ^ 0x0fff) - 1;
    else
        return raw16;
}

/**
 * Receive data from the indicated device, using the given device register, and
 * storing the result in the output parameters provided. The given converter
 * function will convert the raw data to the output form. Return whether the
 * transaction was successful.
 */
bool imu_read_vector(uint8_t device, uint8_t reg, vectorf *out, int16_t (*converter)(uint8_t *))
{
    uint8_t result[6];

    if (!imu_recv(device, reg, result, 6))
        return false;

    out->x = converter(result);
    out->y = converter(result + 2);
    out->z = converter(result + 4);

    return true;
}

bool imu_read_vector_xzy(uint8_t device, uint8_t reg, vectorf *out, int16_t (*converter)(uint8_t *))
{
    double tmp;

    if (imu_read_vector(device, reg, out, converter))
    {
        tmp = out->y;
        out->y = out->z;
        out->z = tmp;
        return true;
    }

    return false;
}

/**
 * Return the given value scaled to the range defined by lower, middle and upper
 * points.
 */
double scale(double value, double lower, double middle, double upper)
{
    if (value < middle)
        return (value - middle) / (middle - lower);
    else
        return (value - middle) / (upper - middle);
}

/**
 * Normalise the given vector based on the minimum and maximum values.
 */
void normalise(vectorf *in, vectorf *vmin, vectorf *vmax, vectorf *out)
{
    out->x = scale(in->x, vmin->x, (vmin->x + vmax->x) / 2, vmax->x);
    out->y = scale(in->y, vmin->y, (vmin->y + vmax->y) / 2, vmax->y);
    out->z = scale(in->z, vmin->z, (vmin->z + vmax->z) / 2, vmax->z);
}

/**
 * Add a measurement to a circular buffer usable for filtering.
 */
void queue(vectorf values[], int *oldest, int length, vectorf *value)
{
    values[*oldest] = *value;
    *oldest = (*oldest + 1) % length;
}

/**
 * Return a filtered measurement using the given values from a circular buffer
 * with the specified oldest measurement index, and with the buffer having the
 * given length. The filtered measurement is placed in the result vector.
 */
void filter(vectorf values[], int oldest, int length, vectorf *result)
{
    int i;

    result->x = 0;
    result->y = 0;
    result->z = 0;

    /* NOTE: For now, a plain average is used. */

    for (i = 0; i < length; i++)
    {
        result->x += values[i].x;
        result->y += values[i].y;
        result->z += values[i].z;
    }

    result->x /= length;
    result->y /= length;
    result->z /= length;
}

/**
 * Populate the given zero base/levels for the device with the given address,
 * using the specified register to obtain measurements, performing the given
 * number of warm-up measurements and then proper readings, with the specified
 * delay between each one.
 */
void calibrate(vectorf *zerolevel, vectorf zerolevels[], int length, uint8_t address, uint8_t reg, uint16_t delay, int16_t (*converter)(uint8_t *))
{
    int reading = 0, index = 0;
    vectorf level;

    zerolevel->x = 0;
    zerolevel->y = 0;
    zerolevel->z = 0;

    while (reading < IMU_CALIBRATION_WARMUP + IMU_CALIBRATION_READINGS)
    {
        if (imu_read_vector(address, reg, &zerolevels[index], converter))
        {
            index = (index + 1) % length;

            /* Ignore the first few readings. */

            if (reading >= IMU_CALIBRATION_WARMUP)
            {
                filter(zerolevels, index, length, &level);
                zerolevel->x += level.x;
                zerolevel->y += level.y;
                zerolevel->z += level.z;
            }

            reading += 1;
        }

        usleep(delay);
    }

    zerolevel->x /= IMU_CALIBRATION_READINGS;
    zerolevel->y /= IMU_CALIBRATION_READINGS;
    zerolevel->z /= IMU_CALIBRATION_READINGS;
}

/* Accelerometer-specific functions. */

void average_filter(vectorf *value, vectorf buffer[], int *index, int length)
{
    queue(buffer, index, length, value);
    filter(buffer, *index, length, value);
}

void noise_filter(vectorf *value, double threshold)
{
    value->x = noise(value->x, threshold);
    value->y = noise(value->y, threshold);
    value->z = noise(value->z, threshold);
}

void apply_acceleration(double acc, double acc_old, double *pos, double *neg, double seconds)
{
    if (acc > 0)
        *pos += (acc + acc_old) / 2 * seconds;
    else
        *neg += (acc + acc_old) / 2 * seconds;
}

void apply_decay(double acc, double* pos, double* neg)
{
    double decay = 1.0;

    if (*pos < -*neg)
    {
        if ((acc < -ACCEL_THRESHOLD) && (*pos > ACCEL_SUM_THRESHOLD))
            decay = VELOCITY_DECAY_SEVERE;
        if (fabs(acc) <= ACCEL_THRESHOLD)
            decay = VELOCITY_DECAY_GRADUAL;
        *neg *= decay;
    }
    if (*pos > -*neg)
    {
        if ((acc > ACCEL_THRESHOLD) && (*neg < -ACCEL_SUM_THRESHOLD))
            decay = VELOCITY_DECAY_SEVERE;
        if (fabs(acc) <= ACCEL_THRESHOLD)
            decay = VELOCITY_DECAY_GRADUAL;
        *pos *= decay;
    }
}

void update_velocity(vectorf *velocity, vectorf *acceleration, vectorf *acceleration_old, vectorf *apos, vectorf *aneg, double seconds)
{
    apply_acceleration(acceleration->x, acceleration_old->x, &(apos->x), &(aneg->x), seconds);
    apply_acceleration(acceleration->y, acceleration_old->y, &(apos->y), &(aneg->y), seconds);
    apply_acceleration(acceleration->z, acceleration_old->z, &(apos->z), &(aneg->z), seconds);

    apply_decay(acceleration->x, &(apos->x), &(aneg->x));
    apply_decay(acceleration->y, &(apos->y), &(aneg->y));
    apply_decay(acceleration->z, &(apos->z), &(aneg->z));

    velocity->x = apos->x + aneg->x;
    velocity->y = apos->y + aneg->y;
    velocity->z = apos->z + aneg->z;
}

void update_displacement(vectorf *displacement, vectorf *velocity, vectorf *velocity_old, double seconds)
{
    displacement->x += (velocity->x + velocity_old->x) / 2 * seconds;
    displacement->y += (velocity->y + velocity_old->y) / 2 * seconds;
    displacement->z += (velocity->z + velocity_old->z) / 2 * seconds;
}
