/*
 * Ben NanoNote communication with the Pololu MinIMU-9 with the L3G4200D 3-axis
 * gyroscope and LSM303DLM accelerometer/magnetometer.
 *
 * http://www.pololu.com/catalog/product/1265
 *
 * Copyright (C) 2013 Paul Boddie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifndef __IMU_H__
#define __IMU_H__

#include <ubb/ubb.h>
#include "i2c.h"
#include "geo.h"

/* Pin assignments:
 *
 * Sniffer  UBB   MinIMU-9
 * -------  ----  --------
 * DAT2     DAT2
 * CD       DAT3  SCL
 * CMD      CMD   SDA
 * VCC      VDD   VIN
 * CLK      CLK   (3V)
 * GND      GND   GND
 * DAT0     DAT0  (1V8)
 * DAT1     DAT1
 *
 * The 3V and 1V8 pins play no role in the communication, but a six pin header
 * exposing these pins on the IMU can be used to connect the IMU with the
 * Sniffer/UBB without any complications.
 */

#define IMU_SCL UBB_DAT3
#define IMU_SDA UBB_CMD
#define IMU_3V  UBB_CLK
#define IMU_1V8 UBB_DAT0

/* I2C addresses and operations. */

#define IMU_GYRO_ADDRESS            0xd2 /* 1101001x */
#define IMU_ACCEL_ADDRESS           0x30 /* 0011000x */
#define IMU_MAGNET_ADDRESS          0x3c /* 0011110x */

/* Gyroscope registers. */

#define IMU_GYRO_WHO_AM_I           0x0f
#define IMU_GYRO_CTRL_REG1          0x20
#define IMU_GYRO_CTRL_REG4          0x23
#define IMU_GYRO_CTRL_REG5          0x24
#define IMU_GYRO_OUT_TEMP           0x26
#define IMU_GYRO_STATUS_REG         0x27

/* Gyroscope register values. */

#define IMU_GYRO_CTRL_REG1_PD       0x08
#define IMU_GYRO_CTRL_REG1_ZEN      0x04
#define IMU_GYRO_CTRL_REG1_YEN      0x02
#define IMU_GYRO_CTRL_REG1_XEN      0x01
#define IMU_GYRO_CTRL_REG1_ALL      0x0f /* composite of power and all axes */
#define IMU_GYRO_CTRL_REG4_BDU      0x80
#define IMU_GYRO_CTRL_REG4_250DPS   0x00 /* FS1 = 0; FS0 = 0 */
#define IMU_GYRO_CTRL_REG4_500DPS   0x10 /* FS1 = 0; FS0 = 1 */
#define IMU_GYRO_CTRL_REG4_2000DPS  0x20 /* FS1 = 1; FS0 = 0 */
#define IMU_GYRO_CTRL_REG5_BOOT     0x80
#define IMU_GYRO_CTRL_REG5_HPEN     0x10
#define IMU_GYRO_CTRL_REG5_OUTSEL1  0x02
#define IMU_GYRO_CTRL_REG5_OUTSEL0  0x01
#define IMU_GYRO_OUT_X_L            0x28
#define IMU_GYRO_OUT_X_H            0x29
#define IMU_GYRO_OUT_Y_L            0x2a
#define IMU_GYRO_OUT_Y_H            0x2b
#define IMU_GYRO_OUT_Z_L            0x2c
#define IMU_GYRO_OUT_Z_H            0x2d
#define IMU_GYRO_READ_MANY          0x80 /* gyroscope register increment */

/* Accelerometer registers. */

#define IMU_ACCEL_CTRL_REG1_A       0x20
#define IMU_ACCEL_CTRL_REG2_A       0x21
#define IMU_ACCEL_CTRL_REG4_A       0x23
#define IMU_ACCEL_HP_FILTER_RESET_A 0x25
#define IMU_ACCEL_OUT_X_L_A         0x28
#define IMU_ACCEL_OUT_X_H_A         0x29
#define IMU_ACCEL_OUT_Y_L_A         0x2a
#define IMU_ACCEL_OUT_Y_H_A         0x2b
#define IMU_ACCEL_OUT_Z_L_A         0x2c
#define IMU_ACCEL_OUT_Z_H_A         0x2d
#define IMU_ACCEL_READ_MANY         0x80 /* accelerometer register increment */

/* Accelerometer register values. */

#define IMU_ACCEL_CTRL_REG1_50HZ    0x20 /* normal power */
#define IMU_ACCEL_CTRL_REG1_100HZ   0x28 /* normal power */
#define IMU_ACCEL_CTRL_REG1_400HZ   0x30 /* normal power */
#define IMU_ACCEL_CTRL_REG1_1000HZ  0x38 /* normal power */
#define IMU_ACCEL_CTRL_REG1_05HZ    0x40 /* low power, 0.5Hz */
#define IMU_ACCEL_CTRL_REG1_1HZ     0x60 /* low power, 1Hz */
#define IMU_ACCEL_CTRL_REG1_2HZ     0x80 /* low power, 2Hz */
#define IMU_ACCEL_CTRL_REG1_5HZ     0x90 /* low power, 5Hz */
#define IMU_ACCEL_CTRL_REG1_10HZ    0xa0 /* low power, 10Hz */
#define IMU_ACCEL_CTRL_REG1_ZEN     0x04
#define IMU_ACCEL_CTRL_REG1_YEN     0x02
#define IMU_ACCEL_CTRL_REG1_XEN     0x01
#define IMU_ACCEL_CTRL_REG1_ALL     0x07 /* composite of all axes */
#define IMU_ACCEL_CTRL_REG2_FDS     0x10
#define IMU_ACCEL_CTRL_REG2_50      0x00 /* filter cut off is rate/50 */
#define IMU_ACCEL_CTRL_REG2_100     0x01 /* filter cut off is rate/100 */
#define IMU_ACCEL_CTRL_REG2_200     0x02 /* filter cut off is rate/200 */
#define IMU_ACCEL_CTRL_REG2_400     0x03 /* filter cut off is rate/400 */
#define IMU_ACCEL_CTRL_REG2_HPCF1   0x02
#define IMU_ACCEL_CTRL_REG4_BDU     0x80
#define IMU_ACCEL_CTRL_REG4_2G      0x00 /* FS1 = 0; FS0 = 0 */
#define IMU_ACCEL_CTRL_REG4_4G      0x10 /* FS1 = 0; FS0 = 1 */
#define IMU_ACCEL_CTRL_REG4_8G      0x30 /* FS1 = 1; FS0 = 1 */

/* Magnetometer registers. */

#define IMU_MAGNET_CRA_REG_M        0x00
#define IMU_MAGNET_CRB_REG_M        0x01
#define IMU_MAGNET_MR_REG_M         0x02
#define IMU_MAGNET_OUT_X_H_M        0x03
#define IMU_MAGNET_OUT_X_L_M        0x04
#define IMU_MAGNET_OUT_Z_H_M        0x05
#define IMU_MAGNET_OUT_Z_L_M        0x06
#define IMU_MAGNET_OUT_Y_H_M        0x07
#define IMU_MAGNET_OUT_Y_L_M        0x08
#define IMU_MAGNET_SR_REG_M         0x09
#define IMU_MAGNET_IRA_REG_M        0x0a
#define IMU_MAGNET_IRB_REG_M        0x0b
#define IMU_MAGNET_IRC_REG_M        0x0c
#define IMU_MAGNET_WHO_AM_I_M       0x0f

/* Magnetometer register values. */

#define IMU_MAGNET_CRA_REG_0_75HZ   0x00 /* 0.75Hz */
#define IMU_MAGNET_CRA_REG_1_5HZ    0x04 /* 1.5Hz */
#define IMU_MAGNET_CRA_REG_3HZ      0x08 /* 3Hz */
#define IMU_MAGNET_CRA_REG_7_5HZ    0x0c /* 7.5Hz */
#define IMU_MAGNET_CRA_REG_15HZ     0x10 /* 15Hz */
#define IMU_MAGNET_CRA_REG_30HZ     0x14 /* 30Hz */
#define IMU_MAGNET_CRA_REG_75HZ     0x18 /* 75Hz */
#define IMU_MAGNET_CRA_REG_220HZ    0x1c /* 220Hz */
#define IMU_MAGNET_CRB_REG_1_3G     0x20 /* 1.3G */
#define IMU_MAGNET_CRB_REG_1_9G     0x40 /* 1.9G */
#define IMU_MAGNET_CRB_REG_2_5G     0x60 /* 2.5G */
#define IMU_MAGNET_CRB_REG_4_0G     0x80 /* 4.0G */
#define IMU_MAGNET_CRB_REG_4_7G     0xa0 /* 4.7G */
#define IMU_MAGNET_CRB_REG_5_6G     0xc0 /* 5.6G */
#define IMU_MAGNET_CRB_REG_8_1G     0xe0 /* 8.1G */
#define IMU_MAGNET_MR_REG_CONT      0x00
#define IMU_MAGNET_MR_REG_SINGLE    0x01
#define IMU_MAGNET_MR_REG_SLEEP     0x02

/* Common values. */

#define IMU_250DPS_UNIT             8750  /* dps -> 8.75mdps */
#define IMU_500DPS_UNIT             17500 /* dps -> 17.5mdps */
#define IMU_2000DPS_UNIT            70000 /* dps -> 70mdps */

#define IMU_2G_UNIT                 1000 /* g -> 1mg */
#define IMU_4G_UNIT                 2000 /* g -> 2mg */
#define IMU_8G_UNIT                 3900 /* g -> 3.9mg */

#define IMU_1_3G_UNIT               909 /* G (1G / 1100) */
#define IMU_1_9G_UNIT               1169 /* G (1G / 855) */
#define IMU_2_5G_UNIT               1492 /* G (1G / 670) */
#define IMU_4_0G_UNIT               2222 /* G (1G / 450) */
#define IMU_4_7G_UNIT               2500 /* G (1G / 400) */
#define IMU_5_6G_UNIT               3030 /* G (1G / 330) */
#define IMU_8_1G_UNIT               4347 /* G (1G / 230) */

#define IMU_MAGNET_Z_XY_RATIO       (8.0 / 9.0) /* adjustments to the above for the smaller Z scale */

/* Measurement parameters. */

#define IMU_UPDATE_PERIOD           10000 /* s -> 10ms == 0.01s (100Hz) */
#define IMU_MAGNET_UPDATE_PERIOD    30000 /* s -> 30ms == 0.03s (33Hz) */
#define TEXT_UPDATE_PERIOD          500000 /* s */
#define GUI_UPDATE_PERIOD           100000 /* s */

#define IMU_CALIBRATION_WARMUP      100
#define IMU_CALIBRATION_READINGS    100

#define IMU_ACCEL_BUFFER_SIZE       10

#define IMU_UDPS_FACTOR             IMU_2000DPS_UNIT
#define IMU_GYRO_DPS_SCALE          IMU_GYRO_CTRL_REG4_2000DPS

#define IMU_UG_FACTOR               IMU_2G_UNIT
#define IMU_ACCEL_SCALE             IMU_ACCEL_CTRL_REG4_2G
#define IMU_ACCEL_FREQ              IMU_ACCEL_CTRL_REG1_50HZ
#define IMU_ACCEL_FILTER_FREQ       IMU_ACCEL_CTRL_REG2_50

#define IMU_UGAUSS_FACTOR           IMU_4_0G_UNIT
#define IMU_MAGNET_SCALE            IMU_MAGNET_CRB_REG_4_0G
#define IMU_MAGNET_FREQ             IMU_MAGNET_CRA_REG_30HZ
#define IMU_MAGNET_SETTINGS_FILE    "imu-magnet.cfg"

#define ACCEL_G                     9.81 /* ms**-2 */

#define ACCEL_THRESHOLD             0.1 /* g */
#define ACCEL_SUM_THRESHOLD         0.01 /* gs */
#define VELOCITY_DECAY_SEVERE       0.5
#define VELOCITY_DECAY_GRADUAL      0.9

#define ACCEL_REST_MAG_LOWER        0.995
#define ACCEL_REST_MAG_UPPER        1.005
#define ROTATION_ADJUSTMENT_FACTOR  0.1

/* Convert microdegrees * microseconds to degrees. */

#define to_angle(x)                 ((double) (x) / ((uint64_t) 1000000000000))

/* Convert microg to g. */

#define to_accel(x)                 ((double) (x) / 1000000)

/* Convert microgauss to gauss. */

#define to_field(x)                 to_accel(x)

/* Get a period in microseconds. */

#define get_period(now, then)       ((now.tv_sec - then.tv_sec) * 1000000 + (now.tv_usec - then.tv_usec))

/* Function definitions. */

bool imu_recv(uint8_t device, uint8_t reg, uint8_t *result, uint8_t len);
bool imu_sendone(uint8_t device, uint8_t reg, uint8_t value);
int16_t convert(uint8_t raw[]);
int16_t convert12(uint8_t raw[]);
int16_t convertBE12L(uint8_t raw[]);
bool imu_read_vector(uint8_t device, uint8_t reg, vectorf *out, int16_t (*converter)(uint8_t *));
bool imu_read_vector_xzy(uint8_t device, uint8_t reg, vectorf *out, int16_t (*converter)(uint8_t *));
double scale(double value, double lower, double middle, double upper);
void normalise(vectorf *in, vectorf *vmin, vectorf *vmax, vectorf *out);
void queue(vectorf values[], int *oldest, int length, vectorf *value);
void filter(vectorf values[], int oldest, int length, vectorf *result);
void calibrate(vectorf *zerolevel, vectorf zerolevels[], int length, uint8_t address, uint8_t reg, uint16_t delay, int16_t (*converter)(uint8_t *));

/* Accelerometer-specific functions. */

void average_filter(vectorf *value, vectorf buffer[], int *index, int length);
void noise_filter(vectorf *value, double threshold);
void apply_acceleration(double acc, double acc_old, double *pos, double *neg, double seconds);
void apply_decay(double acc, double* pos, double* neg);
void update_velocity(vectorf *velocity, vectorf *acceleration, vectorf *acceleration_old, vectorf *apos, vectorf *aneg, double seconds);
void update_displacement(vectorf *displacement, vectorf *velocity, vectorf *velocity_old, double seconds);

/* Function aliases. */

#define imu_init i2c_init

#endif /* __IMU_H__ */
