/*
 * Test Pololu MinIMU-9 measurements.
 *
 * Copyright (C) 2013 Paul Boddie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <pthread.h>
#include "imu.h"
#include "shutdown.h"
#include "gui.h"
#include "measure.h"

int crosses_axis(double current, double previous)
{
    if ((current == 0) && (previous != 0))
        return -sign(previous);
    if ((current != previous) && (sign(current) == -sign(previous)))
        return sign(current);
    return 0;
}

/* Main program. */

int main(int argc, char *argv[])
{
    void *threadresult;
    uint8_t result[6];
    int readings;
    uint32_t period;
    struct timeval now, ui_updated;
    bool paused = false, using_filter = false;

    signal(SIGINT, init_shutdown);

    /* Access the 8:10 port. */

    if (ubb_open(0) < 0) {
        perror("ubb_open");
        return 1;
    }

    ubb_power(1);
    printf("Power on.\n");

    /* Bring the IMU up. */

    imu_init();

    imu_sendone(IMU_GYRO_ADDRESS, IMU_GYRO_CTRL_REG1, IMU_GYRO_CTRL_REG1_ALL);
    imu_sendone(IMU_ACCEL_ADDRESS, IMU_ACCEL_CTRL_REG1_A, IMU_ACCEL_CTRL_REG1_ALL | IMU_ACCEL_CTRL_REG1_50HZ);

    imu_sendone(IMU_GYRO_ADDRESS, IMU_GYRO_CTRL_REG5, 0);

    if ((argc > 1) && (strcmp(argv[1], "-f") == 0))
    {
        imu_sendone(IMU_ACCEL_ADDRESS, IMU_ACCEL_CTRL_REG2_A, IMU_ACCEL_CTRL_REG2_FDS | IMU_ACCEL_FILTER_FREQ);
        imu_recv(IMU_ACCEL_ADDRESS, IMU_ACCEL_HP_FILTER_RESET_A, result, 1);
        using_filter = true;
    }
    else
        imu_sendone(IMU_ACCEL_ADDRESS, IMU_ACCEL_CTRL_REG2_A, 0);

    imu_sendone(IMU_GYRO_ADDRESS, IMU_GYRO_CTRL_REG4, IMU_GYRO_CTRL_REG4_BDU | IMU_GYRO_DPS_SCALE);
    imu_sendone(IMU_ACCEL_ADDRESS, IMU_ACCEL_CTRL_REG4_A, IMU_ACCEL_CTRL_REG4_BDU | IMU_ACCEL_SCALE);

    if (imu_recv(IMU_GYRO_ADDRESS, IMU_GYRO_WHO_AM_I, result, 1))
        printf("Who am I? %x\n", result[0]);

    if (imu_recv(IMU_MAGNET_ADDRESS, IMU_MAGNET_WHO_AM_I_M, result, 1))
        printf("Who am I? %x\n", result[0]);

    if (imu_recv(IMU_MAGNET_ADDRESS, IMU_MAGNET_IRA_REG_M, result, 1))
        printf("Identification A? %x\n", result[0]);

    if (imu_recv(IMU_MAGNET_ADDRESS, IMU_MAGNET_IRB_REG_M, result, 1))
        printf("Identification B? %x\n", result[0]);

    if (imu_recv(IMU_MAGNET_ADDRESS, IMU_MAGNET_IRC_REG_M, result, 1))
        printf("Identification C? %x\n", result[0]);

    gui_init();
    gui_display_init();

    signal(SIGINT, gui_shutdown);

    memset(accelerationB, 0, sizeof(accelerationB));

    ui_calibrate(using_filter, gui_printf, gui_flush);

    memset(accelerationB, 0, sizeof(accelerationB));

    readings = 0;
    gui_clear();
    gui_printf("A0: % 8.1f, % 8.1f, % 8.1f\n", acceleration0.x, acceleration0.y, acceleration0.z);

    /* Create a measurement thread. */

    if (pthread_create(&thread, NULL, get_measurements, &using_filter) != 0)
    {
        perror("pthread_create");
        gui_shutdown(0);
    }

    signal(SIGINT, gui_shutdown_threaded);

    pthread_mutex_init(&mutex, NULL);

    gettimeofday(&ui_updated, NULL);

    while (1)
    {
        gettimeofday(&now, NULL);

        period = get_period(now, ui_updated);

        if ((period >= IMU_UPDATE_PERIOD) && (!paused))
        {
            ui_updated = now;
            readings = (readings + 1) % SCREEN_WIDTH;

            if (readings == 0)
                gui_clear();

            pthread_mutex_lock(&mutex);

            gui_plot(readings % SCREEN_WIDTH, SCREEN_HEIGHT / 2 * (1 - acceleration.x), 255, 0, 0, 255);
            gui_plot(readings % SCREEN_WIDTH, SCREEN_HEIGHT / 2 * (1 - acceleration.y), 0, 255, 0, 255);
            gui_plot(readings % SCREEN_WIDTH, SCREEN_HEIGHT / 2 * (1 - acceleration.z), 0, 0, 255, 255);

            gui_flush();

            printf("A: % 6.4f, % 6.4f, % 6.4f  ",
                   acceleration.x, acceleration.y, acceleration.z);

            printf("DET: % 6.4f, % 6.4f, % 6.4f\n", direction, elevation, tilt);

            pthread_mutex_unlock(&mutex);
        }

        switch (gui_handle_events())
        {
            case IMU_UI_OP_QUIT:
            pthread_cancel(thread);
            pthread_join(thread, &threadresult);
            gui_quit();
            return 0;

            case IMU_UI_OP_PAUSE:
            paused = !paused;
            break;

            default:
            break;
        }

        usleep(IMU_UPDATE_PERIOD);
    }

    /* This should be unreachable. */

    pthread_cancel(thread);
    pthread_join(thread, &threadresult);
    gui_quit();
    return 0;
}
