/*
 * Ben NanoNote communication with the Pololu MinIMU-9 with the L3G4200D 3-axis
 * gyroscope and LSM303DLM accelerometer/magnetometer.
 *
 * http://www.pololu.com/catalog/product/1265
 *
 * Copyright (C) 2013 Paul Boddie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <pthread.h>
#include "imu.h"
#include "shutdown.h"
#include "gui.h"
#include "ui.h"
#include "measure.h"

/* Main program. */

int main(int argc, char *argv[])
{
    int argno = 1;
    void *threadresult;
    uint8_t result[6];
    bool using_filter = false;

    /* Local view state. */

    vectorf _viewx, _viewy, _viewz,
            _accelerationD, _accelerationRD,
            _fieldD, _fieldN;

    /* Timekeeping. */

    struct timeval now, text_updated, gui_updated;

    /* Graphical mode variables. */

    bool graphical = false;
    int (*print)(const char *, ...) = printf;
    void (*flush)() = text_flush;
    void (*clear)() = text_clear;
    void (*quit)() = text_quit;
    void (*shutdown)(int) = text_shutdown;
    void (*shutdown_threaded)(int) = text_shutdown_threaded;
    imu_ui_op (*handle_events)() = text_handle_events;

    memset(accelerationB, 0, sizeof(accelerationB));

    /* Enable a high-pass filter if requested. */

    if ((argc > argno) && (strcmp(argv[argno], "-f") == 0))
    {
        argno++;
        using_filter = true;
    }

    /* Initialise graphical or textual mode. */

    graphical = (argc > argno) && (strcmp(argv[argno], "-g") == 0);

    if (graphical)
    {
        argno++;

        print = gui_printf;
        flush = gui_flush;
        clear = gui_clear;
        quit = gui_quit;
        shutdown = gui_shutdown;
        shutdown_threaded = gui_shutdown_threaded;
        handle_events = gui_handle_events;

        gui_init();
        gui_display_init();
    }

    signal(SIGINT, init_shutdown);

    /* Access the 8:10 port. */

    if (ubb_open(0) < 0) {
        perror("ubb_open");
        return 1;
    }

    ubb_power(1);
    printf("Power on.\n");

    /* Bring the IMU up. */

    imu_init();

    imu_sendone(IMU_GYRO_ADDRESS, IMU_GYRO_CTRL_REG1, IMU_GYRO_CTRL_REG1_ALL);
    imu_sendone(IMU_ACCEL_ADDRESS, IMU_ACCEL_CTRL_REG1_A, IMU_ACCEL_CTRL_REG1_ALL | IMU_ACCEL_FREQ);

    imu_sendone(IMU_GYRO_ADDRESS, IMU_GYRO_CTRL_REG5, 0);

    if (using_filter)
    {
        imu_sendone(IMU_ACCEL_ADDRESS, IMU_ACCEL_CTRL_REG2_A, IMU_ACCEL_CTRL_REG2_FDS | IMU_ACCEL_FILTER_FREQ);
        imu_recv(IMU_ACCEL_ADDRESS, IMU_ACCEL_HP_FILTER_RESET_A, result, 1);
    }
    else
        imu_sendone(IMU_ACCEL_ADDRESS, IMU_ACCEL_CTRL_REG2_A, 0);

    imu_sendone(IMU_GYRO_ADDRESS, IMU_GYRO_CTRL_REG4, IMU_GYRO_CTRL_REG4_BDU | IMU_GYRO_DPS_SCALE);
    imu_sendone(IMU_ACCEL_ADDRESS, IMU_ACCEL_CTRL_REG4_A, IMU_ACCEL_CTRL_REG4_BDU | IMU_ACCEL_SCALE);

    imu_sendone(IMU_MAGNET_ADDRESS, IMU_MAGNET_CRA_REG_M, IMU_MAGNET_FREQ);
    imu_sendone(IMU_MAGNET_ADDRESS, IMU_MAGNET_CRB_REG_M, IMU_MAGNET_SCALE);
    imu_sendone(IMU_MAGNET_ADDRESS, IMU_MAGNET_MR_REG_M, IMU_MAGNET_MR_REG_CONT);

    usleep(IMU_MAGNET_UPDATE_PERIOD);

    if (imu_recv(IMU_GYRO_ADDRESS, IMU_GYRO_WHO_AM_I, result, 1))
        printf("Who am I? %x\n", result[0]);

    if (imu_recv(IMU_MAGNET_ADDRESS, IMU_MAGNET_WHO_AM_I_M, result, 1))
        printf("Who am I? %x\n", result[0]);

    if (imu_recv(IMU_MAGNET_ADDRESS, IMU_MAGNET_IRA_REG_M, result, 1))
        printf("Identification A? %x\n", result[0]);

    if (imu_recv(IMU_MAGNET_ADDRESS, IMU_MAGNET_IRB_REG_M, result, 1))
        printf("Identification B? %x\n", result[0]);

    if (imu_recv(IMU_MAGNET_ADDRESS, IMU_MAGNET_IRC_REG_M, result, 1))
        printf("Identification C? %x\n", result[0]);

    /* Get average values for the gyroscope and accelerometer. */

    if ((argc > argno) && (strcmp(argv[argno], "-c") == 0))
    {
        argno++;
        ui_calibrate(using_filter, print, flush);
    }

    /* Reset the acceleration buffer. */

    memset(accelerationB, 0, sizeof(accelerationB));

    /* Create a measurement thread. */

    if (pthread_create(&thread, NULL, get_measurements, &using_filter) != 0)
    {
        perror("pthread_create");
        shutdown(0);
    }

    signal(SIGINT, shutdown_threaded);

    pthread_mutex_init(&mutex, NULL);

    gettimeofday(&now, NULL);
    text_updated = now; gui_updated = now;

    /* Refresh the display by obtaining measurements made in the measurement
       thread. */

    while (1)
    {
        gettimeofday(&now, NULL);

        if (get_period(now, text_updated) >= TEXT_UPDATE_PERIOD)
        {
            pthread_mutex_lock(&mutex);

            /* Show textual details. */

            if (!graphical)
            {
                clear();
                print("Rotation? %.4f, %.4f, %.4f\n", rotation.x, rotation.y, rotation.z);
                print("Vector? %.4f, %.4f, %.4f\n", viewx.x, viewx.y, viewx.z);
                print("Vector? %.4f, %.4f, %.4f\n", viewy.x, viewy.y, viewy.z);
                print("Vector? %.4f, %.4f, %.4f\n", viewz.x, viewz.y, viewz.z);
                print("Acceleration? %.4f, %.4f, %.4f\n", accelerationD.x, accelerationD.y, accelerationD.z);
                print("Field vector? %.4f, %.4f, %.4f\n", fieldD.x, fieldD.y, fieldD.z);
                flush();
            }

            /* Additional textual output. */

            printf("Period? %d\n", imu_period);
            printf("Direction? %.4f\n", raddeg(direction));
            printf("Elevation? %.4f (%.4f)\n", raddeg(elevation), raddeg(elevationA));
            printf("Tilt? %.4f (%.4f)\n", raddeg(tilt), raddeg(tiltA));
            printf("Acceleration? %.4f, %.4f, %.4f\n", accelerationD.x, accelerationD.y, accelerationD.z);

            pthread_mutex_unlock(&mutex);

            text_updated = now;
        }

        if (graphical && (get_period(now, gui_updated) >= GUI_UPDATE_PERIOD))
        {
            pthread_mutex_lock(&mutex);
            _viewx = viewx;
            _viewy = viewy;
            _viewz = viewz;
            _accelerationD = accelerationD;
            _accelerationRD = accelerationRD;
            _fieldD = fieldD;
            _fieldN = fieldN;
            pthread_mutex_unlock(&mutex);

            clear();
            gui_sky(&_viewx, &_viewy, &_viewz);

            _accelerationRD.x *= 10;
            _accelerationRD.y *= 10;
            _accelerationRD.z *= 10;
            gui_point(&_viewx, &_viewy, &_viewz, &_accelerationRD, 0, 255, 0, 127);
            gui_bar(&_accelerationD);

            _fieldD.x *= 10;
            _fieldD.y *= 10;
            _fieldD.z *= 10;
            gui_point(&_viewx, &_viewy, &_viewz, &_fieldD, 0, 0, 255, 127);

            _fieldN.x *= 10;
            _fieldN.y *= 10;
            _fieldN.z *= 10;
            gui_point(&_viewx, &_viewy, &_viewz, &_fieldN, 255, 0, 0, 127);

            flush();

            gui_updated = now;
        }

        /* Respond to instructions from the UI. */

        switch (handle_events())
        {
            case IMU_UI_OP_QUIT:
            pthread_cancel(thread);
            pthread_join(thread, &threadresult);
            quit();
            return 0;

            case IMU_UI_OP_RESET:
            pthread_mutex_lock(&mutex);
            devicex = devicex0;
            devicey = devicey0;
            devicez = devicez0;
            vectorf_reset(&accelerationD);
            pthread_mutex_unlock(&mutex);
            break;

            case IMU_UI_OP_CALIBRATE:
            pthread_mutex_lock(&mutex);
            clear();
            ui_calibrate(using_filter, print, flush);
            flush();
            memset(accelerationB, 0, sizeof(accelerationB));
            gettimeofday(&imu_updated, NULL);
            pthread_mutex_unlock(&mutex);
            break;

            default:
            break;
        }
    }

    /* This should be unreachable. */

    pthread_cancel(thread);
    pthread_join(thread, &threadresult);
    quit();
    return 0;
}
