/*
 * Ben NanoNote communication with the Pololu MinIMU-9 with the L3G4200D 3-axis
 * gyroscope and LSM303DLM accelerometer/magnetometer.
 *
 * http://www.pololu.com/catalog/product/1265
 *
 * Copyright (C) 2013 Paul Boddie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <pthread.h>
#include "imu.h"
#include "geo.h"

#define __MEASURE_H_PRIVATE__
#include "measure.h"
#undef __MEASURE_H_PRIVATE__

static bool setF0 = false;

/**
 * Perform calibration with feedback given in the user interface.
 */
void ui_calibrate(bool using_filter, int (*print)(const char *, ...), void (*flush)())
{
    vectorf tmpB[1];

    print("Calibrating...\n");
    flush();

    /* Calibrate without a filter for rotation. */

    calibrate(&rotation0, tmpB, 1, IMU_GYRO_ADDRESS, IMU_GYRO_OUT_X_L | IMU_GYRO_READ_MANY, IMU_UPDATE_PERIOD, convert);

    print("Calibrated using (%.4f, %.4f, %.4f).\n", rotation0.x, rotation0.y, rotation0.z);
    flush();

    /* Calibrate using a filter for acceleration. */

    calibrate(&acceleration0, accelerationB, IMU_ACCEL_BUFFER_SIZE,
        IMU_ACCEL_ADDRESS, IMU_ACCEL_OUT_X_L_A | IMU_ACCEL_READ_MANY, IMU_UPDATE_PERIOD, convert12);

    /* Filter out the expected 1g measurement on the z axis. */

    if (!using_filter)
        acceleration0.z += acceleration1g;

    print("Calibrated using (%.4f, %.4f, %.4f).\n", acceleration0.x, acceleration0.y, acceleration0.z);
    flush();

    mag_calibrate();

    print("Calibrated using (%.1f, %.1f, %.1f), (%.1f, %.1f, %.1f).\n",
        fieldmin.x, fieldmin.y, fieldmin.z,
        fieldmax.x, fieldmax.y, fieldmax.z);
}

void mag_calibrate()
{
    FILE *magnetcfg;
    char fieldmins[3][8], fieldmaxs[3][8], *endptr;
    double value;
    int i;

    /* Read magnetometer settings, if possible. */

    magnetcfg = fopen(IMU_MAGNET_SETTINGS_FILE, "r");

    if (magnetcfg != NULL)
    {
        fscanf(magnetcfg, "%7s %7s %7s %7s %7s %7s",
            fieldmins[0], fieldmins[1], fieldmins[2],
            fieldmaxs[0], fieldmaxs[1], fieldmaxs[2]);

        for (i = 0; i < 3; i++)
        {
            value = strtod(fieldmins[i], &endptr);
            if (endptr != fieldmins[i])
                fieldmin.axis[i] = value;
            value = strtod(fieldmaxs[i], &endptr);
            if (endptr != fieldmaxs[i])
                fieldmax.axis[i] = value;
        }

        fclose(magnetcfg);
    }
}

void *get_measurements(void *arg)
{
    struct timeval now;
    uint32_t period;
    bool using_filter = false;
    double accelerationM;
    bool set_reference = false;

    pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);

    if (arg != NULL)
        using_filter = *((bool *) arg);

    /* Initialise the default device orientation. */

    devicex = devicex0;
    devicey = devicey0;
    devicez = devicez0;

    /* Note the time to schedule the next update. */

    gettimeofday(&imu_updated, NULL);
    imu_magnet_updated = imu_updated;

    /* NOTE: Wake up the stupid magnetometer. */

    imu_sendone(IMU_MAGNET_ADDRESS, IMU_MAGNET_MR_REG_M, IMU_MAGNET_MR_REG_CONT);

    /* Actual readings. */

    while (1)
    {
        gettimeofday(&now, NULL);

        period = get_period(now, imu_magnet_updated);

        if (period >= IMU_MAGNET_UPDATE_PERIOD)
        {
            imu_magnet_updated = now;

            pthread_mutex_lock(&mutex);

            if (imu_read_vector_xzy(IMU_MAGNET_ADDRESS, IMU_MAGNET_OUT_X_H_M,
                &field, convertBE12L))
            {
                /* NOTE: Handle stupid magnetometer readings. */

                if (!setF0 && (field.x == 0))
                {
                    field.y = 0; field.z = 0;
                }
                else
                {
                    normalise(&field, &fieldmin, &fieldmax, &field);
                    field.x = to_field(field.x * IMU_UGAUSS_FACTOR);
                    field.y = to_field(field.y * IMU_UGAUSS_FACTOR);
                    field.z = to_field(field.z * IMU_UGAUSS_FACTOR * IMU_MAGNET_Z_XY_RATIO);
                    vectorf_normalise(&field, &field);
                }
            }

            pthread_mutex_unlock(&mutex);
        }

        period = get_period(now, imu_updated);

        if (period >= IMU_UPDATE_PERIOD)
        {
            imu_updated = now;

            pthread_mutex_lock(&mutex);

            imu_period = period;

            if (imu_read_vector(IMU_GYRO_ADDRESS, IMU_GYRO_OUT_X_L | IMU_GYRO_READ_MANY,
                &rotation, convert))
            {
                rotation.x -= rotation0.x;
                rotation.y -= rotation0.y;
                rotation.z -= rotation0.z;

                rotation.x = to_angle(rotation.x * IMU_UDPS_FACTOR * period);
                rotation.y = to_angle(rotation.y * IMU_UDPS_FACTOR * period);
                rotation.z = to_angle(rotation.z * IMU_UDPS_FACTOR * period);

                plane_rotate(&devicey, &devicez, degrad(rotation.x));
                plane_rotate(&devicez, &devicex, degrad(rotation.y));
                plane_rotate(&devicex, &devicey, degrad(rotation.z));
            }

            if (imu_read_vector(IMU_ACCEL_ADDRESS, IMU_ACCEL_OUT_X_L_A | IMU_ACCEL_READ_MANY,
                &acceleration, convert12))
            {
                acceleration.x -= acceleration0.x;
                acceleration.y -= acceleration0.y;
                acceleration.z -= acceleration0.z;

                /* Convert to g. */

                acceleration.x /= acceleration1g;
                acceleration.y /= acceleration1g;
                acceleration.z /= acceleration1g;

                /* Detect gravitational acceleration. */

                accelerationM = vectorf_mag(&acceleration);
                set_reference = (accelerationM > ACCEL_REST_MAG_LOWER) && (accelerationM < ACCEL_REST_MAG_UPPER);

                /* Obtain the acceleration in the global space. */

                vectorf_convert(&acceleration, &devicex, &devicey, &devicez, &accelerationD);
            }

            /* Obtain the view axes and device orientation. */

            vectorf_negate(&devicey, &viewx);
            vectorf_negate(&devicez, &viewy);
            vectorf_negate(&devicex, &viewz);

            direction = vectorf_direction(&viewz);
            elevation = vectorf_elevation(&viewz);
            tilt = within(-(vectorf_tilt_in_plane(&viewy0, &viewx, &viewy) - M_PI / 2), M_PI);

            /* Reset or update the reference acceleration. */

            if (set_reference)
            {
                accelerationRD = accelerationD;
            }
            else
            {
                vectorf_rotate_in_space(&accelerationRD, &viewz, &viewy, &viewx, degrad(rotation.y), &accelerationRD);
                vectorf_rotate_in_space(&accelerationRD, &viewx, &viewy, &viewz, degrad(-rotation.x), &accelerationRD);
            }

            /* Define the tilt and elevation of the reference acceleration. */

            elevationA = vectorf_tilt_in_plane(&accelerationRD, &viewy0, &viewz);
            tiltA = vectorf_tilt_in_plane_with_axis(&accelerationRD, &viewy0, &viewx, &viewz);

            /* Adjust according to elevation. */

            if (set_reference && (fabs(elevation) < degrad(60)))
            {
                plane_rotate(&devicey, &devicez, -tiltA * ROTATION_ADJUSTMENT_FACTOR);
                plane_rotate(&devicez, &devicex, -elevationA * ROTATION_ADJUSTMENT_FACTOR);

                vectorf_negate(&devicey, &viewx);
                vectorf_negate(&devicez, &viewy);
                vectorf_negate(&devicex, &viewz);
            }

            /* Obtain the magnetic field in the global space. */

            if (!vectorf_null(&field))
            {
                directionF = vectorf_direction(&field);
                elevationF = vectorf_elevation(&field);

                /* Define the global vector, remembering the initial value. */

                vectorf_convert(&field, &devicex, &devicey, &devicez, &fieldD);

                if (!setF0)
                {
                    fieldD0 = fieldD;
                    setF0 = true;
                }
            }

            /* Determine the initial field vector in the current device space. */

            if (setF0)
            {
                vectorf_convert_into(&fieldD0, &devicex, &devicey, &devicez, &field0);
                directionF0 = vectorf_direction(&field0);
                elevationF0 = vectorf_elevation(&field0);

                /* Determine the east and north vectors using dynamic field information. */

                vectorf_cross(&viewy0, &fieldD, &fieldE);
                vectorf_normalise(&fieldE, &fieldE);
                vectorf_cross(&fieldE, &viewy0, &fieldN);
            }

            /* Subtract the constant background acceleration. */

            if (!using_filter)
                accelerationD.y -= 1;

            pthread_mutex_unlock(&mutex);
        }

        usleep(IMU_UPDATE_PERIOD);
    }

    return NULL;
}
